from flashkeras.utils.otherimports import *
from flashkeras.utils.kerasimports import *
from flashkeras.utils.typehints import *

class FlashNet:
    def __init__(self, 
                    network: Union[Sequential, Functional],
                    isFullNetwork: bool
                    ) -> None:
        self.network = network
        self.isFullNetwork: bool = isFullNetwork

class FlashTransferLearning:

    def __init__(self,
                 input_shape: tuple[int, int, int] = (-1,-1,-1),
                 include_top: bool = True,
                 weights: str = "imagenet",
                 freeze: bool | int = False,
                 use_only_n_layers: bool | int = False
                 ) -> None:
        
        self.input_shape: tuple[int, int, int] = input_shape
        self.include_top: bool = include_top
        self.weights: str = weights
        self.freeze: bool | int = freeze
        self.use_only_n_layers: bool | int = use_only_n_layers

    def _dropLayers(self, model: Sequential) -> Sequential:
        if type(self.use_only_n_layers) == int:
            if self.use_only_n_layers == 0:
                raise ValueError(
                    "Attribute 'use_only_n_layers' can not be zero, must use at least 1 layer on Transfer Learning."
                )
            elif self.use_only_n_layers > len(model.layers):
                raise ValueError(
                    f"Attribute 'use_only_n_layers' == {self.use_only_n_layers} is higher than the maximum number of layers {len(model.layers)}."
                )
            
        if self.use_only_n_layers == False: 
            return model
        elif self.use_only_n_layers != False and self.include_top:
            print("WARNING: Skipping use_only_n_layers since include_top is True...")
            return model
            
        new_model = Sequential()
        count = 0
        for layer in model.layers:
            if count == self.use_only_n_layers:
                break
            new_model.add(layer)
            count += 1
        
        return new_model
    
    def _freezeLayers(self, model: Functional | Sequential) -> Sequential:
        if not self.freeze: return model

        freezed_model = model

        if type(self.freeze) == bool:
            for layer in freezed_model.layers:
                layer.trainable = False
        else:
            count = 0
            for layer in freezed_model.layers:
                if count < self.freeze: # freezing
                    layer.trainable = False
                count += 1

        return self._dropLayers(freezed_model)

    def transferMyNet(self, path_to_trained_modelh5: str) -> Sequential:
        model = keras.models.load_model(path_to_trained_modelh5)

        new_model = Sequential()

        if self.include_top:
            for layer in model.layers[1:]:
                new_model.add(layer)
        else:
            for layer in model.layers[1:]:
                if isinstance(layer, keras.layers.Flatten):
                    break
                new_model.add(layer)

        for layer in new_model.layers:
            layer.trainable = True
        
        return self._freezeLayers(new_model)

    def transferResnet50(self) -> FlashNet:
        if type(self.use_only_n_layers) == int:
            raise ValueError(
                "ResNet can not have 'use_only_n_layers' attribute as int, use the default value (False)."
            )
        
        if self.include_top:
            print("WARNING: Include top == True. You can not modify this FlashKeras until the flash.clearFlash() method is called.")
            model = self._freezeLayers(ResNet50(include_top=self.include_top,weights=self.weights))
            return FlashNet(model, True)
        
        feature_extractor = ResNet50(
            include_top=self.include_top,
            weights=self.weights,
            input_shape=self.input_shape
        )
        
        if self.freeze == True:
            for layer in feature_extractor.layers:
                layer.trainable = False
        elif type(self.freeze) == int:
            for layer in feature_extractor.layers[:self.freeze]:
                layer.trainable = False

        return FlashNet(feature_extractor, False)
    
    def transferMobileNet(self) -> FlashNet:
        num_layers = len(MobileNet(include_top=False).layers)
        if type(self.use_only_n_layers) == int:
            if int(self.use_only_n_layers) > num_layers:
                raise ValueError(
                    f"Attribute 'use_only_n_layers' == {self.use_only_n_layers} is higher than the maximun number of layers {num_layers}."
                )
        if type(self.freeze) == int:
            if int(self.freeze) > num_layers:
                raise ValueError(
                    f"Attribute 'freeze' == {self.freeze} is higher than the maximun number of layers {num_layers}."
                )

        if self.include_top:
            print("WARNING: Include top == True. You can not modify this FlashKeras until the flash.clearModel() method is called.")
            model = MobileNet(include_top=self.include_top,weights=self.weights)
            return FlashNet(self._freezeLayers(model), True)
        
        feature_extractor = MobileNet(
            include_top=False,
            weights=self.weights,
            input_shape=self.input_shape
        )

        return FlashNet(self._freezeLayers(feature_extractor), False)
    
    def transferXception(self) -> FlashNet:
        num_layers = len(Xception(include_top=False).layers)
        if type(self.use_only_n_layers) == int:
            if int(self.use_only_n_layers) > num_layers:
                raise ValueError(
                    f"Attribute 'use_only_n_layers' == {self.use_only_n_layers} is higher than the maximun number of layers {num_layers}."
                )
        if type(self.freeze) == int:
            if int(self.freeze) > num_layers:
                raise ValueError(
                    f"Attribute 'freeze' == {self.freeze} is higher than the maximun number of layers {num_layers}."
                )
        
        if self.include_top:
            print("WARNING: Include top == True. You can not modify this FlashKeras until the flash.clearModel() method is called.")
            model = Xception(include_top=self.include_top,weights=self.weights)
            return FlashNet(self._freezeLayers(model), True)
        
        feature_extractor = Xception(
            include_top=False,
            weights=self.weights,
            input_shape=self.input_shape
        )

        return FlashNet(self._freezeLayers(feature_extractor), False)
    
    def transferVGG16(self) -> FlashNet:
        num_layers = len(VGG16(include_top=False).layers)
        if type(self.use_only_n_layers) == int:
            if int(self.use_only_n_layers) > num_layers:
                raise ValueError(
                    f"Attribute 'use_only_n_layers' == {self.use_only_n_layers} is higher than the maximun number of layers {num_layers}."
                )
        if type(self.freeze) == int:
            if int(self.freeze) > num_layers:
                raise ValueError(
                    f"Attribute 'freeze' == {self.freeze} is higher than the maximun number of layers {num_layers}."
                )
        
        if self.include_top:
            print("WARNING: Include top == True. You can not modify this FlashKeras until the flash.clearModel() method is called.")
            model = VGG16(include_top=self.include_top,weights=self.weights)
            return FlashNet(self._freezeLayers(model), True)
        
        feature_extractor = VGG16(
            include_top=False,
            weights=self.weights,
            input_shape=self.input_shape
        )

        return FlashNet(self._freezeLayers(feature_extractor), False)




from flashkeras.utils.kerasimports import Sequential

def print_model_summary(model: Sequential):
    # Header
    print(f"{'Index':<5} | {'Layer Type':<24} {'(trainable)':<11} | Configuration")
    print("-" * 80)

    # Iterate through model layers
    for idx, layer in enumerate(model.layers):
        layer_type = layer.__class__.__name__
        config = layer.get_config()
        trainable = "(True)" if layer.trainable else "(freezed)"

        if layer_type == 'Dense':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Units: {config['units']}, Activation: {config['activation']}")

        elif layer_type == 'Conv2D':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Filters: {config['filters']}, Kernel Size: {config['kernel_size']}, Activation: {config['activation']}")

        elif layer_type == 'DepthwiseConv2D':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Depth Multiplier: {config['depth_multiplier']}, Kernel Size: {config['kernel_size']}, Activation: {config['activation']}")

        elif layer_type == 'BatchNormalization':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Epsilon: {config['epsilon']}, Momentum: {config['momentum']}")

        elif layer_type == 'Dropout':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Rate: {config['rate']}")

        elif layer_type == 'ReLU' or 'activation' in config:
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Activation: {layer_type.lower()}")

        elif layer_type == 'Flatten':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | None ")

        elif layer_type == 'MaxPooling2D':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Pool Size: {config['pool_size']}")

        elif layer_type == 'AveragePooling2D':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Pool Size: {config['pool_size']}")

        elif layer_type == 'GlobalMaxPooling2D':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Global Max Pooling")

        elif layer_type == 'GlobalAveragePooling2D':
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | Global Average Pooling")

        else:
            print(f"{idx:<5} | {layer_type:<25}{trainable:<11} | --- ")


from typing import overload, Union, Tuple, Literal, Any, Optional, cast
from keras.preprocessing.image import DirectoryIterator, NumpyArrayIterator # type: ignore
BatchIterator = Union[DirectoryIterator, NumpyArrayIterator]


from flashkeras.utils.otherimports import *
from flashkeras.utils.kerasimports import *
from flashkeras.utils.typehints import *
from flashkeras.utils.filesutils import count_directories_in_directory
from flashkeras.preprocessing.FlashPreProcessing import FlashPreProcessing as prepro

class FlashDataGenerator:
    """ Loads and preprocess images from directories, numpy arrays
    
    """
    def __init__(self,         
                    img_shape: tuple[int, int],
                    color_mode: Literal["rgb", "grayscale"] = "rgb",
                    horizontal_flip: bool = False,
                    rotation_range: int = 0,
                    zoom_range: float = 0,
                    brightness_range: tuple[float, float] | None = None,
                    fill_mode: str = "nearest"
                ) -> None:
    
        self.img_shape: tuple[int, int] = img_shape
        self.color_mode: Literal["rgb", "grayscale"] = color_mode
        self.horizontal_flip: bool = horizontal_flip
        self.rotation_range: int = rotation_range
        self.zoom_range: float = zoom_range
        self.brightness_range: tuple[float, float] | None = brightness_range
        self.fill_mode: str = fill_mode

    def _getClassMode(self, path_or_class_list: str | list[str]):
        num_classes: int = 0
        if type(path_or_class_list) == str:
            num_classes = count_directories_in_directory(path_or_class_list)
        else:
            num_classes = len(path_or_class_list)

        class_mode: str = ""
        if num_classes == 2:
            class_mode = "binary"
        elif num_classes > 2 or num_classes == 1:
            class_mode = "categorical"
        else:
            raise ValueError ("Invalid number of classes!.")

        return class_mode

    def flow_images_from_directory(self,
                                   directory_path: str,
                                   batch_size: int = 32
                                   ) -> DirectoryIterator:
        
        data_gen = ImageDataGenerator(
                    rescale=1./255,
                    horizontal_flip=self.horizontal_flip,
                    rotation_range=self.rotation_range,
                    zoom_range=self.zoom_range,
                    brightness_range=self.brightness_range,
                    fill_mode=self.fill_mode
                )

        image_iterator = data_gen.flow_from_directory(
            directory_path,
            color_mode=self.color_mode,
            target_size=self.img_shape,
            batch_size=batch_size,
            class_mode=None,
            shuffle=False
        )

        return image_iterator

    def flow_classes_from_nparray(self, 
                                  x: np.ndarray, 
                                  y: np.ndarray | None = None,
                                  batch_size: int = 32
                                  ) -> NumpyArrayIterator:
        
        data_gen = ImageDataGenerator(
                    rescale=1./255,
                    horizontal_flip=self.horizontal_flip,
                    rotation_range=self.rotation_range,
                    zoom_range=self.zoom_range,
                    brightness_range=self.brightness_range,
                    fill_mode=self.fill_mode
                )
        
        if self.color_mode == 'rgb':
            x = prepro.convertNdArrayToRGB(x)
        if self.color_mode == 'grayscale':
            x = prepro.convertNdArrayToGrayScale(x)

        x = prepro.resizeNpArray(x, self.img_shape[0], self.img_shape[1])
        
        batches = data_gen.flow(x, y, batch_size, shuffle=True)

        return batches
    
    def flow_classes_from_nparray_test_split(self, 
                                  x: np.ndarray, 
                                  y: np.ndarray,
                                  test_split: float = 0.2,
                                  batch_size: int = 32
                                  ) -> tuple[NumpyArrayIterator, NumpyArrayIterator]:
        
        data_gen = ImageDataGenerator(
                    rescale=1./255,
                    validation_split=test_split,
                    horizontal_flip=self.horizontal_flip,
                    rotation_range=self.rotation_range,
                    zoom_range=self.zoom_range,
                    brightness_range=self.brightness_range,
                    fill_mode=self.fill_mode
                )
        
        if self.color_mode == 'rgb':
            x = prepro.convertNdArrayToRGB(x)
        if self.color_mode == 'grayscale':
            x = prepro.convertNdArrayToGrayScale(x)
    
        x = prepro.resizeNpArray(x, self.img_shape[0], self.img_shape[1])
        
        y = prepro.ensureOneHotEncoding(y)
        
        train_batches = data_gen.flow(x, y, batch_size, subset='training', shuffle=True)

        test_batches = data_gen.flow(x, y, batch_size, subset='validation', shuffle=True)

        return (train_batches, test_batches) 

    def flow_all_classes_from_dir(
            self,
            path_to_main_dir: str, 
            batch_size: int = 32
            ) -> DirectoryIterator | None:
        
        class_mode = self._getClassMode(path_to_main_dir)
        
        data_gen = ImageDataGenerator(
                    rescale=1./255,
                    horizontal_flip=self.horizontal_flip,
                    rotation_range=self.rotation_range,
                    zoom_range=self.zoom_range,
                    brightness_range=self.brightness_range,
                    fill_mode=self.fill_mode
                )

        batches = data_gen.flow_from_directory(
            path_to_main_dir,
            color_mode=self.color_mode,
            target_size=self.img_shape,
            class_mode=class_mode,
            batch_size=batch_size,
            shuffle=True,
        )

        return batches
    
    def flow_all_classes_from_dir_test_split(
            self,
            path_to_main_dir: str,
            test_split: float = 0.2, 
            batch_size: int = 32
            ) -> tuple[DirectoryIterator, DirectoryIterator] | None:

        class_mode = self._getClassMode(path_to_main_dir)

        data_gen = ImageDataGenerator(
                    rescale=1./255,
                    validation_split=test_split,
                    horizontal_flip=self.horizontal_flip,
                    rotation_range=self.rotation_range,
                    zoom_range=self.zoom_range,
                    brightness_range=self.brightness_range,
                    fill_mode=self.fill_mode
                )

        train_batches = data_gen.flow_from_directory(
            path_to_main_dir,
            color_mode=self.color_mode,
            target_size=self.img_shape,
            class_mode=class_mode,
            batch_size=batch_size,
            shuffle=True,
            subset='training'
        )

        test_batches = data_gen.flow_from_directory(
            path_to_main_dir,
            color_mode=self.color_mode,
            target_size=self.img_shape,
            class_mode=class_mode,
            batch_size=batch_size,
            shuffle=True,
            subset='validation'
        )

        return train_batches, test_batches

    def flow_classes_from_dir(
            self,
            path_to_main_dir: str, 
            classes: list[str], 
            batch_size: int = 32
            ) -> DirectoryIterator | None:
        
        class_mode = self._getClassMode(classes)
        
        data_gen = ImageDataGenerator(
                    rescale=1./255,
                    horizontal_flip=self.horizontal_flip,
                    rotation_range=self.rotation_range,
                    zoom_range=self.zoom_range,
                    brightness_range=self.brightness_range,
                    fill_mode=self.fill_mode
                )

        batches = data_gen.flow_from_directory(
            path_to_main_dir,
            color_mode=self.color_mode,
            target_size=self.img_shape,
            class_mode=class_mode,
            classes=classes,
            batch_size=batch_size,
            shuffle=True,
        )

        return batches

    def flow_classes_from_dir_test_split(
            self,
            path_to_main_dir: str, 
            classes: list[str],
            test_split: float = 0.2, 
            batch_size: int = 32
            ) -> tuple[DirectoryIterator, DirectoryIterator] | None:
        
        class_mode: str = self._getClassMode(classes)
        
        data_gen = ImageDataGenerator(
                    rescale=1./255,
                    validation_split=test_split,
                    horizontal_flip=self.horizontal_flip,
                    rotation_range=self.rotation_range,
                    zoom_range=self.zoom_range,
                    brightness_range=self.brightness_range,
                    fill_mode=self.fill_mode
                )

        train_batches = data_gen.flow_from_directory(
            path_to_main_dir,
            color_mode=self.color_mode,
            target_size=self.img_shape,
            class_mode=class_mode,
            classes=classes,
            batch_size=batch_size,
            shuffle=True,
            subset='training'
        )

        test_batches = data_gen.flow_from_directory(
            path_to_main_dir,
            color_mode=self.color_mode,
            target_size=self.img_shape,
            class_mode=class_mode,
            classes=classes,
            batch_size=batch_size,
            shuffle=True,
            subset='validation'
        )

        return train_batches, test_batches


from flashkeras.utils.otherimports import *
from flashkeras.utils.kerasimports import *
from flashkeras.utils.typehints import *
from flashkeras.models import FlashSequential 
from sklearn.metrics import accuracy_score, precision_score, recall_score, roc_auc_score, mean_squared_error, mean_absolute_error, f1_score, auc, roc_curve # type: ignore

class FlashEvaluating:

    @staticmethod
    def _adjustClassMetrics(model: FlashSequential | Sequential, 
                            x_test: pd.DataFrame | np.ndarray | BatchIterator, 
                            y_test: pd.Series | np.ndarray | None = None
                            ) -> tuple[pd.Series | np.ndarray, Any]:

        if not isinstance(model, Sequential):
            true_model = model.model
        else:
            true_model = model

        if isinstance(x_test, (DirectoryIterator, NumpyArrayIterator)):
            x_list = [x for x, _ in x_test]
            y_list = [y for _, y in x_test]
            x_test = np.concatenate(x_list, axis=0)
            y_test = np.concatenate(y_list, axis=0)

            y_pred = true_model.predict(x_test)
            y_pred_classes = y_pred.argmax(axis=-1)

            if y_test is None: 
                raise ValueError("``y_test`` became None at some point.")

            if y_test.ndim > 1 and y_test.shape[1] > 1:
                y_test = np.argmax(y_test, axis=-1)

            if y_test is None: 
                raise ValueError("``y_test`` became None at ``np.argmax(y_test, axis=-1)``.")

            return y_test, y_pred_classes

        if y_test is None:
            raise ValueError('``y_test`` must be provided if Test Data is not a ``BatchIterator``')

        if isinstance(x_test, pd.DataFrame):
            x_test = x_test.values

        y_pred = true_model.predict(x_test)
        y_pred_classes = y_pred.argmax(axis=-1)

        if isinstance(y_test, pd.Series):
            y_test = y_test.to_numpy()

        if y_test.ndim > 1 and y_test.shape[1] > 1:
            y_test = np.argmax(y_test, axis=-1)

        if y_test is None: 
            raise ValueError("``y_test`` became None at ``np.argmax(y_test, axis=-1)``.")

        return y_test, y_pred_classes

    @staticmethod
    @overload
    def getAccuracy(model: FlashSequential | Sequential, x_test: pd.DataFrame | np.ndarray, y_test: pd.Series | np.ndarray | None) -> float: ...
    @staticmethod
    @overload
    def getAccuracy(model: FlashSequential | Sequential, x_test: BatchIterator) -> float: ...
    @staticmethod
    def getAccuracy(model: FlashSequential | Sequential, 
                    x_test: pd.DataFrame | np.ndarray | BatchIterator, 
                    y_test: pd.Series | np.ndarray | None = None
                    ) -> float:
        
        y_test, y_pred_classes = FlashEvaluating._adjustClassMetrics(model, x_test, y_test)
        
        return accuracy_score(y_test, y_pred_classes)

    @staticmethod
    @overload
    def getPrecision(model: FlashSequential | Sequential, x_test: pd.DataFrame | np.ndarray, y_test: pd.Series | np.ndarray) -> float: ...
    @staticmethod
    @overload
    def getPrecision(model: FlashSequential | Sequential, x_test: BatchIterator) -> float: ...
    @staticmethod
    def getPrecision(model: FlashSequential | Sequential, 
                    x_test: pd.DataFrame | np.ndarray | BatchIterator, 
                    y_test: pd.Series | np.ndarray | None = None
                    ) -> float:
        
        y_test, y_pred_classes = FlashEvaluating._adjustClassMetrics(model, x_test, y_test)
        
        return precision_score(y_test, y_pred_classes, average='macro')
    
    @staticmethod
    @overload
    def getRecall(model: FlashSequential | Sequential, x_test: pd.DataFrame | np.ndarray, y_test: pd.Series | np.ndarray) -> float: ...
    @staticmethod
    @overload
    def getRecall(model: FlashSequential | Sequential, x_test: BatchIterator) -> float: ...
    @staticmethod
    def getRecall(model: FlashSequential | Sequential, 
                    x_test: pd.DataFrame | np.ndarray | BatchIterator, 
                    y_test: pd.Series | np.ndarray | None = None
                    ) -> float:
        
        y_test, y_pred_classes = FlashEvaluating._adjustClassMetrics(model, x_test, y_test)
        
        return recall_score(y_test, y_pred_classes, average='macro')

    @staticmethod
    def getMSE(model: FlashSequential | Sequential, x_test: pd.DataFrame | np.ndarray, y_test: pd.Series | np.ndarray):
        if not isinstance(model, Sequential):
            true_model = model.model
        else:
            true_model = model

        y_pred = true_model.predict(x_test)
        return mean_squared_error(y_test, y_pred)

    @staticmethod
    def getMAE(model: FlashSequential | Sequential, x_test: pd.DataFrame | np.ndarray, y_test: pd.Series | np.ndarray):
        if not isinstance(model, Sequential):
            true_model = model.model
        else:
            true_model = model

        y_pred = true_model.predict(x_test)

        return mean_absolute_error(y_test, y_pred)


from keras.datasets import cifar10, cifar100, fashion_mnist, mnist # type: ignore
from sklearn.datasets import load_digits # type: ignore

def load_cifar10():
    """Loads the CIFAR10 dataset.

    This is a dataset of 50,000 32x32 color training images and 10,000 test
    images, labeled over 10 categories. See more info at the
    [CIFAR homepage](https://www.cs.toronto.edu/~kriz/cifar.html).

    The classes are:

    | Label | Description |
    |:-----:|-------------|
    |   0   | airplane    |
    |   1   | automobile  |
    |   2   | bird        |
    |   3   | cat         |
    |   4   | deer        |
    |   5   | dog         |
    |   6   | frog        |
    |   7   | horse       |
    |   8   | ship        |
    |   9   | truck       |

    Returns:
      Tuple of NumPy arrays: `(x_train, y_train), (x_test, y_test)`.

    **x_train**: uint8 NumPy array of image data with shapes
      `(50000, 32, 32, 3)`, containing the training data. Pixel values range
      from 0 to 255.

    **y_train**: uint8 NumPy array of labels (integers in range 0-9)
      with shape `(50000, 1)` for the training data.

    **x_test**: uint8 NumPy array of image data with shapes
      `(10000, 32, 32, 3)`, containing the test data. Pixel values range
      from 0 to 255.

    **y_test**: uint8 NumPy array of labels (integers in range 0-9)
      with shape `(10000, 1)` for the test data.

    Example:

    ```python
    (x_train, y_train), (x_test, y_test) = keras.datasets.cifar10.load_data()
    assert x_train.shape == (50000, 32, 32, 3)
    assert x_test.shape == (10000, 32, 32, 3)
    assert y_train.shape == (50000, 1)
    assert y_test.shape == (10000, 1)
    ```
    """
    return cifar10.load_data()

def load_cifar100():
    """Loads the CIFAR100 dataset.

    This is a dataset of 50,000 32x32 color training images and
    10,000 test images, labeled over 100 fine-grained classes that are
    grouped into 20 coarse-grained classes. See more info at the
    [CIFAR homepage](https://www.cs.toronto.edu/~kriz/cifar.html).

    Args:
      label_mode: one of "fine", "coarse". If it is "fine" the category labels
        are the fine-grained labels, if it is "coarse" the output labels are the
        coarse-grained superclasses.

    Returns:
      Tuple of NumPy arrays: `(x_train, y_train), (x_test, y_test)`.

    **x_train**: uint8 NumPy array of image data with shapes
      `(50000, 32, 32, 3)`, containing the training data. Pixel values range
      from 0 to 255.

    **y_train**: uint8 NumPy array of labels (integers in range 0-99)
      with shape `(50000, 1)` for the training data.

    **x_test**: uint8 NumPy array of image data with shapes
      `(10000, 32, 32, 3)`, containing the test data. Pixel values range
      from 0 to 255.

    **y_test**: uint8 NumPy array of labels (integers in range 0-99)
      with shape `(10000, 1)` for the test data.

    Example:

    ```python
    (x_train, y_train), (x_test, y_test) = keras.datasets.cifar100.load_data()
    assert x_train.shape == (50000, 32, 32, 3)
    assert x_test.shape == (10000, 32, 32, 3)
    assert y_train.shape == (50000, 1)
    assert y_test.shape == (10000, 1)
    ```
    """
    return cifar100.load_data()

def load_fashion_mnist():
    """Loads the Fashion-MNIST dataset.
    This is a dataset of 60,000 28x28 grayscale images of 10 fashion categories,
    along with a test set of 10,000 images. This dataset can be used as
    a drop-in replacement for MNIST.
    The classes are:
    | Label | Description |
    |:-----:|-------------|
    |   0   | T-shirt/top |
    |   1   | Trouser     |
    |   2   | Pullover    |
    |   3   | Dress       |
    |   4   | Coat        |
    |   5   | Sandal      |
    |   6   | Shirt       |
    |   7   | Sneaker     |
    |   8   | Bag         |
    |   9   | Ankle boot  |
    Returns:
    Tuple of NumPy arrays: `(x_train, y_train), (x_test, y_test)`.
    **x_train**: uint8 NumPy array of grayscale image data with shapes
    `(60000, 28, 28)`, containing the training data.
    **y_train**: uint8 NumPy array of labels (integers in range 0-9)
    with shape `(60000,)` for the training data.
    **x_test**: uint8 NumPy array of grayscale image data with shapes
    (10000, 28, 28), containing the test data.
    **y_test**: uint8 NumPy array of labels (integers in range 0-9)
    with shape `(10000,)` for the test data.
    Example:
    ```python
    (x_train, y_train), (x_test, y_test) = fashion_mnist.load_data()
    assert x_train.shape == (60000, 28, 28)
    assert x_test.shape == (10000, 28, 28)
    assert y_train.shape == (60000,)
    assert y_test.shape == (10000,)
    ```
    License:
    The copyright for Fashion-MNIST is held by Zalando SE.
    Fashion-MNIST is licensed under the [MIT license](
    https://github.com/zalandoresearch/fashion-mnist/blob/master/LICENSE).
    """
    return fashion_mnist.load_data()

def load_mnist():
    """Loads the MNIST dataset.

    This is a dataset of 60,000 28x28 grayscale images of the 10 digits,
    along with a test set of 10,000 images.
    More info can be found at the
    [MNIST homepage](http://yann.lecun.com/exdb/mnist/).

    Args:
      path: path where to cache the dataset locally
        (relative to `~/.keras/datasets`).

    Returns:
      Tuple of NumPy arrays: `(x_train, y_train), (x_test, y_test)`.

    **x_train**: uint8 NumPy array of grayscale image data with shapes
      `(60000, 28, 28)`, containing the training data. Pixel values range
      from 0 to 255.

    **y_train**: uint8 NumPy array of digit labels (integers in range 0-9)
      with shape `(60000,)` for the training data.

    **x_test**: uint8 NumPy array of grayscale image data with shapes
      (10000, 28, 28), containing the test data. Pixel values range
      from 0 to 255.

    **y_test**: uint8 NumPy array of digit labels (integers in range 0-9)
      with shape `(10000,)` for the test data.

    Example:

    ```python
    (x_train, y_train), (x_test, y_test) = keras.datasets.mnist.load_data()
    assert x_train.shape == (60000, 28, 28)
    assert x_test.shape == (10000, 28, 28)
    assert y_train.shape == (60000,)
    assert y_test.shape == (10000,)
    ```

    License:
      Yann LeCun and Corinna Cortes hold the copyright of MNIST dataset,
      which is a derivative work from original NIST datasets.
      MNIST dataset is made available under the terms of the
      [Creative Commons Attribution-Share Alike 3.0 license.](
      https://creativecommons.org/licenses/by-sa/3.0/)
    """
    return mnist.load_data()


from flashkeras.preprocessing.FlashPreProcessing import FlashPreProcessing as preprocess
from flashkeras.models.FlashTransferLearning import FlashNet
from flashkeras.analysing.models import print_model_summary
from flashkeras.utils.otherimports import *
from flashkeras.utils.kerasimports import *
from flashkeras.utils.typehints import *

tasks_available = Literal['classification', 'regression']

class FlashSequential:
    def __init__(self, task: Literal['classification', 'regression']) -> None:
        self.task: Literal['classification', 'regression'] = task

        self.model: Sequential = Sequential()
        self.layers = self.model.layers
        self.blocked: list[str] = []

        self.output_activation: Literal["sigmoid", "softmax"] = "sigmoid"
        self.output_loss: str = "binary_crossentropy"
        self.output_neurons: int = 1   

        self.optimizer: Any = None
        self.metrics: Any = None       

    def add(self, layer) -> None:       
        self._checkBlocked()

        self.model.add(layer)
        
        self.layers = self.model.layers

    def addTransferLearning(self, transferLayer: FlashNet) -> None: 
        '''
            Adds transfer learning layers created with FlashTransferLearning
        '''

        self._checkBlocked()

        if transferLayer.isFullNetwork: 
            self.model = transferLayer.network
            self.blocked.append('TransferLearning: Include Top was set. To maintain integrity of the flash, any modifycations to the architecture are blocked.')
        elif isinstance(transferLayer.network, Sequential):
            for l in transferLayer.network.layers:
                self.model.add(l)
        else:
            self.model.add(transferLayer.network)   

    def compile(self,    
                optimizer: str | Any = "adam",
                learning_rate: float | None = None,
                loss: Any | None = None,
                metrics: Any = None,
                ) -> None:

        if metrics is None:
            if self.task == 'classification':
                metrics = ['accuracy']
            elif self.task == 'regression':
                metrics = ['mae']
        
        if isinstance(optimizer, str):
            opt = self._optimizerMap(optimizer, learning_rate)
        else:
            opt = optimizer

        self.model.compile(opt, loss, metrics)

        self.optimizer = opt
        self.metrics = metrics

    def build(self, 
              data: Union[np.ndarray, pd.DataFrame, BatchIterator], 
              y: Union[np.ndarray, pd.DataFrame, pd.Series, None] = None,
              auto_output_layer: bool = True
              ) -> None:
        ''' Automatically sets the `input_shape` and `output_layer` based on your data
        '''
        self._setOutputParams(y, data)

        if not self.model.inputs:
            self.setInputShape(preprocess.getInputShape(data))

        if auto_output_layer:

            if self.task == 'classification':
                self.model.add(Dense(self.output_neurons, self.output_activation))
            elif self.task == 'regression':
                self.model.add(Dense(self.output_neurons))

    def fit(self, 
            x: Any | None = None,
            y: Any | None = None,
            epochs: int = 1,
            auto_output_layer: bool = False,
            validation_data: Any | None = None,
            steps_per_epoch: Any | None = None,
            batch_size: Any | None = None,
            verbose: str = "auto",
            callbacks: Any | None = None,
            validation_split: float = 0,
            shuffle: bool = True,
            class_weight: Any | None = None,
            sample_weight: Any | None = None,
            initial_epoch: int = 0
            ) -> Any:
        
        self.build(x, y, auto_output_layer)

        if not self.model._is_compiled:
            self.compile(loss=self.output_loss)
        
        if not self.model.loss:
            self.compile(optimizer=self.model.optimizer,
                         loss=self.output_loss,
                         metrics=self.metrics)

        if isinstance(x, (DirectoryIterator, NumpyArrayIterator)):
            history = self.model.fit(x, epochs=epochs, validation_data=validation_data, steps_per_epoch=steps_per_epoch,
                           batch_size=batch_size, verbose=verbose, callbacks=callbacks, validation_split=validation_split,
                           shuffle=shuffle, class_weight=class_weight, sample_weight=sample_weight, initial_epoch=initial_epoch)
            return history
        
        if x is not None and y is not None:
            history = self.model.fit(x, y, epochs=epochs, validation_data=validation_data, steps_per_epoch=steps_per_epoch,
                           batch_size=batch_size, verbose=verbose, callbacks=callbacks, validation_split=validation_split,
                           shuffle=shuffle, class_weight=class_weight, sample_weight=sample_weight, initial_epoch=initial_epoch)
            return history
        elif x is not None and y is None:
            history = self.model.fit(x, epochs=epochs, validation_data=validation_data, steps_per_epoch=steps_per_epoch,
                           batch_size=batch_size, verbose=verbose, callbacks=callbacks, validation_split=validation_split,
                           shuffle=shuffle, class_weight=class_weight, sample_weight=sample_weight, initial_epoch=initial_epoch)
            return history
        
    def predict(self,
                x: Any,
                batch_size: Any | None = None,
                verbose: str = "auto",
                steps: Any | None = None,
                # callbacks: Any | None = None,
                # max_queue_size: int = 10,
                # workers: int = 1,
                # use_multiprocessing: bool = False
                ) -> Any:
        
        return self.model.predict(x=x, batch_size=batch_size, verbose=verbose, steps=steps)

    def singlePredict(self, instance: Any):
        if isinstance(instance, np.ndarray):
            if len(instance.shape) == 2:
                instance = np.expand_dims(instance, axis=-1)

            if len(instance.shape) == 3:
                instance = np.expand_dims(instance, axis=0)
            elif len(instance.shape) == 1:
                instance = np.expand_dims(instance, axis=0)

        elif isinstance(instance, pd.DataFrame):
            if len(instance.shape) == 1 or instance.shape[0] == 1:
                instance = instance.to_numpy().reshape(1, -1)

        prediction = cast(np.ndarray, self.predict(instance))

        if len(prediction.shape) == 2:
            return np.argmax(prediction, axis=1)
        
        elif len(prediction.shape) == 1:
            return prediction

        return prediction

    def summary(self):
        try:
            self.model.summary()
        except:
            print("WARNING: The model is not built yet. Only the architecture will be shown.")
            print_model_summary(self.model)

    def loadModel(self, path_to_modelh5: str):
        print("This will Overwrite an existent model.")
        self.model = keras.models.load_model(path_to_modelh5)
        self.blocked.append('Loaded Model: You have loaded a full model. To maintain integrity of the flash, any modifications to the architecture are blocked.')

    def setInputShape(self, input_shape: tuple):
        new_model: Sequential = Sequential(InputLayer(input_shape=input_shape))
        for layer in self.model.layers: 
            new_model.add(layer)
        self.model = new_model
            
    def clearFlash(self) -> None:
        '''
            Reset every configuration made on flash model.
        '''
        self.model = Sequential()
        self.layers = self.model.layers
        self.blocked = []

    def _optimizerMap(self, opt: str, lr: Optional[float]):
        if lr is None:
            lr = 0.0001
        if opt == "adam":
            return Adam(learning_rate=lr)
        elif opt == "nadam":
            return Nadam(learning_rate=lr)
        elif opt == "sgd":
            return SGD(learning_rate=lr)
        else:
            return Adam()
        
    def _setOutputParams(self,
                y: Union[np.ndarray, pd.DataFrame, pd.Series, None] = None, 
                image_batches: Optional[BatchIterator] = None, 
                ) -> None:
        
        if self.task == 'classification':
            sparse_or_not: str = ""
            if y is not None:
                if isinstance(y, (pd.Series, np.ndarray)):
                    if y.ndim == 1:
                        sparse_or_not += "sparse_"
                        num_classes = len(np.unique(y))
                    elif y.ndim == 2:
                        num_classes = y.shape[1]
            
            elif image_batches is not None:
                if isinstance(image_batches, DirectoryIterator):
                    num_classes = image_batches.num_classes
                elif isinstance(image_batches, NumpyArrayIterator):
                    num_classes = len(image_batches.y[0])
            else:
                raise ValueError("Either x or y must be provided.")
            
            # Determine activation and loss
            if num_classes == 2:
                self.output_activation = "sigmoid"
                self.output_loss = "binary_crossentropy"
                self.output_neurons = 1
            else:
                self.output_activation = "softmax"
                self.output_loss = sparse_or_not + "categorical_crossentropy"
                self.output_neurons = num_classes
        
        elif self.task == 'regression':
            if y is None:
                raise ValueError('``y`` cannot be None value.')
            
            if isinstance(y, (pd.Series, np.ndarray)):
                if y.ndim > 1:
                    self.output_neurons = y.shape[1]
                else:
                    self.output_neurons = 1
            else:
                self.output_neurons = len(y.columns)
            
            self.output_loss = 'mse'

    def _checkBlocked(self) -> None:
        if len(self.blocked) != 0:
            err_message = 'Possible errors:\n'
            for error in self.blocked:
                err_message = err_message + error + '\n'

            raise ValueError(err_message)



from flashkeras.utils.otherimports import *
from flashkeras.utils.typehints import *
from flashkeras.utils.typehints import BatchIterator

def show_images_from_batch(batches: BatchIterator, num_images: int = 1, fig_size: tuple[int, int] = (15,5)):
    all_images = []
    images, labels = (None, None)
    while True:
        tuple_or_images = next(batches)
        try:
            images, labels = tuple_or_images
        except:
            images = tuple_or_images

        for img in images:
            all_images.append(img)
        if batches.batch_index == 0: break
    
    num_images = min(num_images, len(all_images))
    
    plt.figure(figsize=fig_size)
    
    for i in range(num_images):
        plt.subplot(1, num_images, i + 1)
        plt.imshow(all_images[i])
        plt.axis('off')
    
    plt.show()

def show_images_nparray(images: np.ndarray, num_images: int = 1, fig_size: tuple[int, int] = (15,5)):
    num_images = min(num_images, len(images))

    plt.figure(figsize=fig_size)
    
    for i in range(num_images):
        plt.subplot(1, num_images, i+1)
        plt.imshow(images[i])
        plt.axis('off')

    plt.show()

def show_image_nparray(image: np.ndarray, fig_size: tuple[int, int] = (7,7)):
    plt.figure(figsize=fig_size)
    
    plt.imshow(image)
    plt.axis('off')

    plt.show()

def show_images_from_directory(dir_path: str, num_images=1):

    files = os.listdir(dir_path)
    
    valid_extentions = ['.jpg', '.jpeg', '.png', '.bmp', '.gif']
    images = [f for f in files if os.path.splitext(f)[1].lower() in valid_extentions]
    
    num_images = min(num_images, len(images))
    
    plt.figure(figsize=(15, 5))
    
    for i in range(num_images):
        img_path = os.path.join(dir_path, images[i])
        img = mpimg.imread(img_path)
        plt.subplot(1, num_images, i + 1)
        plt.imshow(img)
        plt.axis('off')
    
    plt.show()

def show_image(image: Union[np.ndarray, Image.Image, str], fig_size: tuple[int, int] = (7,7)) -> Tuple[int, int]:
    '''Provide the image or path to the image and shows it.
    '''

    if isinstance(image, str):
        image = Image.open(image)

    if isinstance(image, Image.Image):
        image = np.array(image)

    if isinstance(image, np.ndarray):
        show_image_nparray(image, fig_size)
    
    raise ValueError("The image must be one of the following types: ``np.ndarray``, ``Image.Image`` or a ``str`` representing the path.")



from .layers import *

_exclude = [
    'AvgPool1D', 
    'AvgPool2D', 
    'AvgPool3D',
    'AbstractRNNCell',
    'ActivityRegularization',
    'Add',
    'add',
    'AdditiveAttention',
    'Attention',
    'Average',
    'average',
    'concatenate',
    'Cropping1D',
    'Cropping2D',
    'Cropping3D',
    'deserialize',
    'Dot',
    'dot',
    'GlobalAvgPool1D',
    'GlobalAvgPool2D',
    'GlobalAvgPool3D',
    'GlobalMaxPool1D',
    'GlobalMaxPool2D',
    'GlobalMaxPool3D',
    'Wrapper',
    'experimental',
    'maximum',
    'subtract',
    'multiply',
    'minimum',
    'serialize'
]

__all__ = [name for name in dir() if name not in _exclude and not name.startswith('_')]


from keras.preprocessing.image import ImageDataGenerator # type: ignore
from keras.models import Sequential # type: ignore
from keras.layers import InputLayer, Dense # type: ignore
from keras.optimizers import Adam, Nadam, SGD # type: ignore
import keras # type: ignore
from keras.applications import MobileNet, ResNet50, Xception, VGG16 # type: ignore
from keras.src.engine.functional import Functional # type: ignore
import tensorflow as tf # type: ignore
from keras.utils import to_categorical # type: ignore
from keras.preprocessing.image import img_to_array, load_img # type: ignore


import numpy as np # type: ignore
import pandas as pd # type: ignore
import cv2 # type: ignore
import matplotlib.pyplot as plt # type: ignore
import os # type: ignore
import matplotlib.image as mpimg # type: ignore
from sklearn.preprocessing import MinMaxScaler # type: ignore
from sklearn.model_selection import train_test_split # type: ignore
from sklearn.preprocessing import LabelEncoder # type: ignore
from PIL import Image # type: ignore


from .images import show_image_nparray, show_images_from_batch, show_images_from_directory, show_images_nparray
from .models import print_model_summary

__all__ = [
    'show_image_nparray', 
    'show_images_from_batch', 
    'show_images_from_directory', 
    'show_images_nparray', 
    'print_model_summary'
]


from .models.FlashSequential import FlashSequential
from .models.FlashTransferLearning import FlashTransferLearning
from .preprocessing.FlashPreProcessing import FlashPreProcessing
from .preprocessing.images.FlashDataGenerator import FlashDataGenerator
from .evaluation.FlashEvaluating import FlashEvaluating
import pandas
import numpy

__all__ = [
           'FlashSequential',
           'FlashTransferLearning',
           'FlashPreProcessing',
           'FlashDataGenerator',
           'FlashEvaluating'
]


from flashkeras.utils.otherimports import *
from flashkeras.utils.typehints import *
from sklearn.metrics import mean_squared_error # type: ignore

def plot_line_graph(y_values: list | np.ndarray, 
                    x_values: list | Any = None,
                    graph_title='Line Graph',
                    x_label='X', 
                    y_label='Y',
                    x_rotation=90,
                    fig_size=(15, 6),
                    x_ticks=None,
                    y_ticks=None,
                    grid: bool = False
                    ):

    if x_values is None:
        x_values = np.arange(1, len(y_values) + 1)

    # Plot the original data points
    plt.figure(figsize=fig_size)
    plt.plot(x_values, y_values, marker='o', linestyle='-', color='b', label=x_label)

    # Add titles and labels
    plt.title(graph_title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    if x_ticks is None:
        x_ticks = x_values
    plt.xticks(x_ticks, rotation=x_rotation)

    if y_ticks is None:
        y_min, y_max = min(y_values), max(y_values)
        y_ticks = np.linspace(y_min, y_max, 20)
    plt.yticks(y_ticks)

    # Disable grid, add legend, and show plot
    plt.grid(grid)
    plt.legend()
    plt.show()

def get_line_graph(y_values: list | np.ndarray, 
                   x_values: list | Any = None,
                   graph_title='Line Graph',
                   x_label='X', 
                   y_label='Y',
                   x_rotation=90,
                   fig_size=(15, 6),
                   x_ticks=None,
                   y_ticks=None,
                   grid: bool = False):
    
    if x_values is None:
        x_values = np.arange(1, len(y_values) + 1)

    fig, ax = plt.subplots(figsize=fig_size)
    
    ax.plot(x_values, y_values, marker='o', linestyle='-', color='b', label=x_label)

    ax.set_title(graph_title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if x_ticks is None:
        x_ticks = x_values
    ax.set_xticks(x_ticks)
    ax.tick_params(axis='x', rotation=x_rotation)

    if y_ticks is None:
        y_min, y_max = min(y_values), max(y_values)
        y_ticks = np.linspace(y_min, y_max, 20)
    ax.set_yticks(y_ticks)

    ax.grid(grid)
    ax.legend()

    return fig

def plot_multi_line_graph(coords: list[list[list]], 
                          x_values: list = None,
                          graph_title='Line Graph',
                          x_label='X', 
                          y_label='Y',
                          line_labels: list | None | Literal['default'] = None, 
                          x_rotation=90,
                          fig_size=(15, 6),
                          x_ticks=None,
                          y_ticks=None,
                          grid: bool = False):
    """Plots multiple lines on the same graph with the given coordnates
    
    :param coords: List of coordnates lists i.e. [ [[x, y], [x, y]], [[x, y], [x, y], [x, y]] ].
    :param labels: Names of the different lines.
    """
    
    if x_values is None:
        x_values = [list(zip(*coord))[0] for coord in coords]
    
    y_values = [list(zip(*coord))[1] for coord in coords]

    plt.figure(figsize=fig_size)

    if line_labels == 'default':
        line_labels = [f'Line {i+1}' for i in range(len(coords))]

    for i, (x, y) in enumerate(zip(x_values, y_values)):
        if line_labels is not None:
            plt.plot(x, y, marker='o', linestyle='-', label=line_labels[i])
        else:
            plt.plot(x, y, marker='o', linestyle='-')

    plt.title(graph_title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    if x_ticks is None:
        x_ticks = np.concatenate(x_values)
    plt.xticks(x_ticks, rotation=x_rotation)

    if y_ticks is None:
        all_y_values = np.concatenate(y_values)
        y_min, y_max = min(all_y_values), max(all_y_values)
        y_ticks = np.linspace(y_min, y_max, 20)
    plt.yticks(y_ticks)

    plt.grid(grid)
    plt.legend()

    plt.show()

def get_multi_line_graph(coords: list[list[list]], 
                         x_values: list = None,
                         graph_title='Line Graph',
                         x_label='X', 
                         y_label='Y',
                         line_labels: list | None | Literal['default'] = None, 
                         x_rotation=90,
                         fig_size=(15, 6),
                         x_ticks=None,
                         y_ticks=None,
                         grid: bool = False):
    """Generate a multi-lines graph with the given coordnates and returns it
    
    :param coords: List of coordnates lists i.e. [ [[x, y], [x, y]], [[x, y], [x, y], [x, y]] ].
    :param labels: Names of the different lines.
    """
    
    if x_values is None:
        x_values = [list(zip(*coord))[0] for coord in coords]
    
    y_values = [list(zip(*coord))[1] for coord in coords]

    fig, ax = plt.subplots(figsize=fig_size)

    if line_labels == 'default':
        line_labels = [f'Line {i+1}' for i in range(len(coords))]

    for i, (x, y) in enumerate(zip(x_values, y_values)):
        if line_labels is not None:
            plt.plot(x, y, marker='o', linestyle='-', label=line_labels[i])
        else:
            plt.plot(x, y, marker='o', linestyle='-')

    ax.set_title(graph_title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if x_ticks is None:
        x_ticks = np.concatenate(x_values)
    ax.set_xticks(x_ticks)
    ax.tick_params(axis='x', rotation=x_rotation)

    if y_ticks is None:
        all_y_values = np.concatenate(y_values)
        y_min, y_max = min(all_y_values), max(all_y_values)
        y_ticks = np.linspace(y_min, y_max, 20)
    ax.set_yticks(y_ticks)

    ax.grid(grid)
    ax.legend()

    return fig

def plot_multi_line_graph_polynomial_fit(coords: list[list[list]], 
                          max_degree=5,
                          graph_title='Line Graph',
                          x_label='X', 
                          y_label='Y',
                          line_labels: list | None | Literal['default'] = None, 
                          x_rotation=90,
                          fig_size=(15, 6),
                          x_ticks=None,
                          y_ticks=None,
                          grid: bool = False,
                          regression_line_color='black'
                          ):
    """Plots multiple lines on the same graph with the given coordnates and also plots the polynomial regression of all the points.
    
    :param coords: List of coordnates lists i.e. [ [[x, y], [x, y]], [[x, y], [x, y], [x, y]] ].
    :param labels: Names of the different lines.
    """
    
    def find_best_polynomial_degree(x, y, max_degree):
        best_degree = 1
        best_error = float('inf')

        for degree in range(1, max_degree + 1):
            coefficients = np.polyfit(x, y, degree)
            polynomial = np.poly1d(coefficients)
            y_pred = polynomial(x)
            error = mean_squared_error(y, y_pred)

            if error < best_error:
                best_error = error
                best_degree = degree

        return best_degree
    
    x_values_flat = np.concatenate([list(zip(*coord))[0] for coord in coords])
    y_values_flat = np.concatenate([list(zip(*coord))[1] for coord in coords])

    plt.figure(figsize=fig_size)

    best_degree = find_best_polynomial_degree(x_values_flat, y_values_flat, max_degree)

    coefficients = np.polyfit(x_values_flat, y_values_flat, best_degree)
    polynomial = np.poly1d(coefficients)
    regression_y_values = polynomial(np.sort(x_values_flat))

    if line_labels == 'default':
        line_labels = [f'Line {i+1}' for i in range(len(coords))]

    for i, coord in enumerate(coords):
        x, y = zip(*coord)
        if line_labels is not None:
            plt.plot(x, y, marker='o', linestyle='-', label=line_labels[i])
        else:
            plt.plot(x, y, marker='o', linestyle='-')

    plt.plot(np.sort(x_values_flat), regression_y_values, linewidth=3, linestyle='--', color=regression_line_color, label=f'Regression (Degree {best_degree})')

    plt.title(graph_title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    if x_ticks is None:
        x_ticks = np.sort(np.unique(x_values_flat))
    plt.xticks(x_ticks, rotation=x_rotation)

    if y_ticks is None:
        y_min, y_max = min(y_values_flat), max(y_values_flat)
        y_ticks = np.linspace(y_min, y_max, 20)
    plt.yticks(y_ticks)

    plt.grid(grid)
    plt.legend()

    plt.show()

def get_multi_line_graph_polynomial_fit(coords: list[list[list]], 
                          max_degree=5,
                          graph_title='Line Graph',
                          x_label='X', 
                          y_label='Y',
                          line_labels: list | None | Literal['default'] = None, 
                          x_rotation=90,
                          fig_size=(15, 6),
                          x_ticks=None,
                          y_ticks=None,
                          grid: bool = False):
    """Generate a multi-lines graph with the given coordnates and a polynomial regression and returns it
    
    :param coords: List of coordnates lists i.e. [ [[x, y], [x, y]], [[x, y], [x, y], [x, y]] ].
    :param labels: Names of the different lines.
    """
    
    def find_best_polynomial_degree(x, y, max_degree):
        best_degree = 1
        best_error = float('inf')

        for degree in range(1, max_degree + 1):
            coefficients = np.polyfit(x, y, degree)
            polynomial = np.poly1d(coefficients)
            y_pred = polynomial(x)
            error = mean_squared_error(y, y_pred)

            if error < best_error:
                best_error = error
                best_degree = degree

        return best_degree
    
    x_values_flat = np.concatenate([list(zip(*coord))[0] for coord in coords])
    y_values_flat = np.concatenate([list(zip(*coord))[1] for coord in coords])

    fig, ax = plt.subplots(figsize=fig_size)

    best_degree = find_best_polynomial_degree(x_values_flat, y_values_flat, max_degree)

    coefficients = np.polyfit(x_values_flat, y_values_flat, best_degree)
    polynomial = np.poly1d(coefficients)
    regression_y_values = polynomial(np.sort(x_values_flat))

    if line_labels == 'default':
        line_labels = [f'Line {i+1}' for i in range(len(coords))]

    for i, coord in enumerate(coords):
        x, y = zip(*coord)
        if line_labels is not None:
            plt.plot(x, y, marker='o', linestyle='-', label=line_labels[i])
        else:
            plt.plot(x, y, marker='o', linestyle='-')

    ax.plot(np.sort(x_values_flat), regression_y_values, linewidth=3, linestyle='--', color='b', label=f'Regression (Degree {best_degree})')

    ax.set_title(graph_title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if x_ticks is None:
        x_ticks = np.sort(np.unique(x_values_flat))
    ax.set_xticks(x_ticks)
    ax.tick_params(axis='x', rotation=x_rotation)

    if y_ticks is None:
        y_min, y_max = min(y_values_flat), max(y_values_flat)
        y_ticks = np.linspace(y_min, y_max, 20)
    ax.set_yticks(y_ticks)

    ax.grid(grid)
    ax.legend()

    return fig

def plot_line_graph_polynomial_fit_zoom_higher_y(y_values: list | np.ndarray, 
                                               x_values: list | Any = None,
                                               max_degree=5, 
                                               focus_threshold=0.8,
                                               graph_title='Graph with Polynomial Regression & Y axis Cut',
                                               x_label='X', 
                                               y_label='Y',
                                               x_rotation=90,
                                               fig_size=(15, 6),
                                               x_ticks=None,
                                               grid: bool = False
                                               ) -> None:

    def find_y_limits(y_values) -> tuple[float, float]:
        highest_value = max(y_values)
        lower_limit = highest_value * focus_threshold
        lower_dif = highest_value
        lower_value = lower_limit
        for v in y_values:
            dif = v - lower_limit
            if dif < lower_dif and dif >= 0:
                lower_value = v
                lower_dif = dif
        
        lower_value = lower_value - (lower_value / 150)
        highest_value = highest_value + (highest_value / 150)

        return lower_value, highest_value

    def find_best_polynomial_degree(x, y, max_degree):
        best_degree = 1
        best_error = float('inf')

        for degree in range(1, max_degree + 1):
            coefficients = np.polyfit(x, y, degree)
            polynomial = np.poly1d(coefficients)
            y_pred = polynomial(x)
            error = mean_squared_error(y, y_pred)

            if error < best_error:
                best_error = error
                best_degree = degree

        return best_degree

    if x_values is None:
        x_values = np.arange(1, len(y_values) + 1)

    best_degree = find_best_polynomial_degree(x_values, y_values, max_degree)

    coefficients = np.polyfit(x_values, y_values, best_degree)
    polynomial = np.poly1d(coefficients)
    regression_y_values = polynomial(x_values)

    plt.figure(figsize=fig_size)
    
    plt.plot(x_values, y_values, marker='o', linestyle='-', color='b', label=x_label)

    plt.plot(x_values, regression_y_values, linewidth=3, linestyle='--', color='r', label='Regression')

    plt.title(graph_title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    if x_ticks is None:
        x_ticks = x_values
        
    plt.xticks(x_ticks, rotation=x_rotation)

    min_val, max_val = find_y_limits(y_values)
    plt.ylim(min_val, max_val)

    plt.grid(grid)
    plt.legend()
    plt.show()

def get_line_graph_polynomial_fit_zoom_higher_y(y_values: list | np.ndarray, 
                                                x_values: list | Any = None,
                                                max_degree=5, 
                                                focus_threshold=0.8,
                                                graph_title='Graph with Polynomial Regression & Y axis Cut',
                                                x_label='X', 
                                                y_label='Y',
                                                x_rotation=90,
                                                fig_size=(15, 6),
                                                x_ticks=None,
                                                grid: bool = False) -> plt.Figure:
    
    def find_y_limits(y_values) -> tuple[float, float]:
        highest_value = max(y_values)
        lower_limit = highest_value * focus_threshold
        lower_dif = highest_value
        lower_value = lower_limit
        for v in y_values:
            dif = v - lower_limit
            if dif < lower_dif and dif >= 0:
                lower_value = v
                lower_dif = dif
        
        lower_value = lower_value - (lower_value / 150)
        highest_value = highest_value + (highest_value / 150)

        return lower_value, highest_value

    def find_best_polynomial_degree(x, y, max_degree):
        best_degree = 1
        best_error = float('inf')

        for degree in range(1, max_degree + 1):
            coefficients = np.polyfit(x, y, degree)
            polynomial = np.poly1d(coefficients)
            y_pred = polynomial(x)
            error = mean_squared_error(y, y_pred)

            if error < best_error:
                best_error = error
                best_degree = degree

        return best_degree

    if x_values is None:
        x_values = np.arange(1, len(y_values) + 1)

    best_degree = find_best_polynomial_degree(x_values, y_values, max_degree)

    coefficients = np.polyfit(x_values, y_values, best_degree)
    polynomial = np.poly1d(coefficients)
    regression_y_values = polynomial(x_values)

    fig, ax = plt.subplots(figsize=fig_size)
    
    ax.plot(x_values, y_values, marker='o', linestyle='-', color='b', label=x_label)

    ax.plot(x_values, regression_y_values, linewidth=3, linestyle='--', color='r', label='Regression')

    ax.set_title(graph_title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if x_ticks is None:
        x_ticks = x_values
    ax.set_xticks(x_ticks)
    ax.tick_params(axis='x', rotation=x_rotation)

    min_val, max_val = find_y_limits(y_values)
    ax.set_ylim(min_val, max_val)

    ax.grid(grid)
    ax.legend()

    return fig

def plot_line_graph_polynomial_fit(y_values: list | np.ndarray, 
                                   x_values: list | Any = None,
                                   max_degree=5, 
                                   graph_title='Graph with Polynomial Regression',
                                   x_label='X', 
                                   y_label='Y',
                                   x_rotation=90,
                                   fig_size=(15, 6),
                                   x_ticks=None,
                                   y_ticks=None,
                                   grid: bool = False
                                   ):
    
    def find_best_polynomial_degree(x, y, max_degree):
        best_degree = 1
        best_error = float('inf')

        for degree in range(1, max_degree + 1):
            coefficients = np.polyfit(x, y, degree)
            polynomial = np.poly1d(coefficients)
            y_pred = polynomial(x)
            error = mean_squared_error(y, y_pred)

            if error < best_error:
                best_error = error
                best_degree = degree

        return best_degree

    if x_values is None:
        x_values = np.arange(1, len(y_values) + 1)

    # Find the best polynomial degree
    best_degree = find_best_polynomial_degree(x_values, y_values, max_degree)

    # Calculate the best polynomial regression
    coefficients = np.polyfit(x_values, y_values, best_degree)
    polynomial = np.poly1d(coefficients)
    regression_y_values = polynomial(x_values)

    # Plot the original data points
    plt.figure(figsize=fig_size)
    plt.plot(x_values, y_values, marker='o', linestyle='-', color='b', label=x_label)

    # Plot the best fit regression line
    plt.plot(x_values, regression_y_values, linewidth=3, linestyle='--', color='r', label='Regression')

    # Add titles and labels
    plt.title(graph_title)
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    if x_ticks is None:
        x_ticks = x_values

    plt.xticks(x_ticks, rotation=x_rotation)

    if y_ticks is None:
        y_min, y_max = min(y_values), max(y_values)
        y_ticks = np.linspace(y_min, y_max, 20)
    plt.yticks(y_ticks)

    plt.grid(grid)
    plt.legend()
    plt.show()

def get_line_graph_polynomial_fit(y_values: list | np.ndarray, 
                                  x_values: list = None,
                                  max_degree=5, 
                                  graph_title='Graph with Polynomial Regression',
                                  x_label='X', 
                                  y_label='Y',
                                  x_rotation=90,
                                  fig_size=(7, 4),  # Tamanho da figura ajustado para ser mais apropriado para Tkinter
                                  x_ticks=None,
                                  y_ticks=None,
                                  grid: bool = False):

    def find_best_polynomial_degree(x, y, max_degree):
        best_degree = 1
        best_error = float('inf')

        for degree in range(1, max_degree + 1):
            coefficients = np.polyfit(x, y, degree)
            polynomial = np.poly1d(coefficients)
            y_pred = polynomial(x)
            error = mean_squared_error(y, y_pred)

            if error < best_error:
                best_error = error
                best_degree = degree

        return best_degree

    if x_values is None:
        x_values = np.arange(1, len(y_values) + 1)

    best_degree = find_best_polynomial_degree(x_values, y_values, max_degree)

    coefficients = np.polyfit(x_values, y_values, best_degree)
    polynomial = np.poly1d(coefficients)
    regression_y_values = polynomial(x_values)

    fig, ax = plt.subplots(figsize=fig_size)
    
    ax.plot(x_values, y_values, marker='o', linestyle='-', color='b', label='Dados Originais')

    ax.plot(x_values, regression_y_values, linewidth=3, linestyle='--', color='r', label='Regressão Polinomial')

    # Define título e rótulos
    ax.set_title(graph_title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    if x_ticks is None:
        x_ticks = x_values
    ax.set_xticks(x_ticks)
    ax.tick_params(axis='x', rotation=x_rotation)

    if y_ticks is None:
        y_min, y_max = min(y_values), max(y_values)
        y_ticks = np.linspace(y_min, y_max, 20)
    ax.set_yticks(y_ticks)

    ax.grid(grid)
    ax.legend()

    return fig



from setuptools import setup, find_packages # type: ignore

setup(
    name="flashkeras",
    version="0.1.4",
    packages=find_packages(include=['flashkeras', 'flashkeras.*']),
    install_requires=[ 
        'tensorflow==2.15.0',
        'matplotlib>=3.9.2',
        'opencv-python>=4.10.0.84',
        'pandas>=2.2.2',
        'scikit-learn>=1.5.1'
    ],
    author="diego dias fernandes",
    author_email="diegodiasfernandes.comp@gmail.com",
    description="Package to fasten Keras manipulation and overall machine learning pipeline.",
    url="https://github.com/diegodiasfernandes/flashkeras",
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10,<3.12',
)



from .dataframes import load_wine, load_breast_cancer, load_diabetes, load_iris, boston_housing_load_data
from .image import load_cifar100, load_mnist, load_cifar10, load_fashion_mnist, load_digits
from .matrix_like import load_imdb, imdb_get_world_index, load_reuters, reuters_get_word_index, reuters_get_label_names

__all__ = [
    'load_wine', 
    'load_breast_cancer', 
    'load_diabetes', 
    'load_iris', 
    'load_cifar100',
    'load_mnist',
    'load_cifar10',
    'load_fashion_mnist',
    'load_digits',
    'load_imdb',
    'imdb_get_world_index',
    'load_reuters',
    'reuters_get_word_index',
    'reuters_get_word_index',
    'reuters_get_label_names',
    'boston_housing_load_data'
]


import os

def count_directories_in_directory(path: str) -> int:
    if not os.path.isdir(path):
        raise ValueError(f"The path '{path}' is not a valid directory.")
    
    return sum(1 for item in os.listdir(path) if os.path.isdir(os.path.join(path, item)))


from flashkeras.utils.otherimports import *
from flashkeras.utils.kerasimports import *
from flashkeras.utils.typehints import *

class FlashPreProcessing:

    @staticmethod
    def train_test_split(*arrays: Any, 
                         test_split: float | None = None,
                         random_state: int | None = None,
                         ) -> list:
        """Split arrays or matrices into random train and test subsets.

        Quick utility that wraps input validation,
        ``next(ShuffleSplit().split(X, y))``, and application to input data
        into a single call for splitting (and optionally subsampling) data into a
        one-liner.

        Read more in the :ref:`User Guide <cross_validation>`.

        Parameters
        ----------
        *arrays : sequence of indexables with same length / shape[0]
            Allowed inputs are lists, numpy arrays, scipy-sparse
            matrices or pandas dataframes.

        test_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the
            absolute number of test samples. If None, the value is set to the
            complement of the train size. If ``train_size`` is also None, it will
            be set to 0.25.

        train_size : float or int, default=None
            If float, should be between 0.0 and 1.0 and represent the
            proportion of the dataset to include in the train split. If
            int, represents the absolute number of train samples. If None,
            the value is automatically set to the complement of the test size.

        random_state : int, RandomState instance or None, default=None
            Controls the shuffling applied to the data before applying the split.
            Pass an int for reproducible output across multiple function calls.
            See :term:`Glossary <random_state>`.

        shuffle : bool, default=True
            Whether or not to shuffle the data before splitting. If shuffle=False
            then stratify must be None.

        stratify : array-like, default=None
            If not None, data is split in a stratified fashion, using this as
            the class labels.
            Read more in the :ref:`User Guide <stratification>`.

        Returns
        -------
        splitting : list, length=2 * len(arrays)
            List containing train-test split of inputs.

            .. versionadded:: 0.16
                If the input is sparse, the output will be a
                ``scipy.sparse.csr_matrix``. Else, output type is the same as the
                input type.

        Examples
        --------
        >>> import numpy as np
        >>> from sklearn.model_selection import train_test_split
        >>> X, y = np.arange(10).reshape((5, 2)), range(5)
        >>> X
        array([[0, 1],
            [2, 3],
            [4, 5],
            [6, 7],
            [8, 9]])
        >>> list(y)
        [0, 1, 2, 3, 4]

        >>> x_train, x_test, y_train, y_test = train_test_split(
        ...     X, y, test_size=0.33, random_state=42)
        ...
        >>> x_train
        array([[4, 5],
            [0, 1],
            [6, 7]])
        >>> y_train
        [2, 0, 3]
        >>> x_test
        array([[2, 3],
            [8, 9]])
        >>> y_test
        [1, 4]

        >>> train_test_split(y, shuffle=False)
        [[0, 1, 2], [3, 4]]
        """
        return train_test_split(*arrays, test_size=test_split, random_state=random_state)

    @staticmethod
    def stackDataFrames(matrix_a: pd.DataFrame | pd.Series | np.ndarray, 
                        matrix_b: pd.DataFrame | pd.Series | np.ndarray
                        ) -> pd.DataFrame | pd.Series | np.ndarray:
        
        def to_2d_array(data):
            if isinstance(data, pd.DataFrame):
                return data.values
            elif isinstance(data, pd.Series):
                return data.values.reshape(-1, 1)
            elif isinstance(data, np.ndarray):
                if data.ndim == 1:
                    return data.reshape(-1, 1)
                return data
            else:
                raise ValueError("Inputs must be either pd.DataFrame, pd.Series, or np.ndarray")

        matrix_a_2d = to_2d_array(matrix_a)
        matrix_b_2d = to_2d_array(matrix_b)

        if matrix_a_2d.shape[1] != matrix_b_2d.shape[1]:
            raise ValueError(f"Shape mismatch: matrix_a has shape {matrix_a_2d.shape} and matrix_b has shape {matrix_b_2d.shape}. Both must have the same number of columns.")

        stacked = np.vstack((matrix_a_2d, matrix_b_2d))

        if isinstance(matrix_a, pd.DataFrame) and isinstance(matrix_b, pd.DataFrame):
            return pd.DataFrame(stacked, columns=matrix_a.columns)
        
        elif isinstance(matrix_a, pd.Series) and isinstance(matrix_b, pd.Series):
            return pd.Series(stacked.flatten(), name=matrix_a.name)
        
        elif isinstance(matrix_a, np.ndarray) and isinstance(matrix_b, np.ndarray):
            if matrix_a.ndim == 1 and matrix_b.ndim == 1:
                return stacked.flatten()
            return stacked 
        
        return stacked

    @overload
    @staticmethod
    def stackImageDatasets(data_a: np.ndarray, data_b: np.ndarray) -> np.ndarray: ...
    
    @overload
    @staticmethod
    def stackImageDatasets(data_a: Tuple[np.ndarray, np.ndarray], data_b: Tuple[np.ndarray, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]: ...
    
    @staticmethod
    def stackImageDatasets(data_a: np.ndarray | Tuple[np.ndarray, np.ndarray], 
                       data_b: np.ndarray | Tuple[np.ndarray, np.ndarray]
                       ) -> np.ndarray | Tuple[np.ndarray, np.ndarray]:
        
        if isinstance(data_a, np.ndarray) and isinstance(data_b, np.ndarray):
            if data_a.shape[1:] != data_b.shape[1:]:
                raise ValueError(f"Shape mismatch: data_a has shape {data_a.shape} and data_b has shape {data_b.shape}. "
                                 f"Both must have the same shape except for the first dimension.")

            return np.concatenate([data_a, data_b], axis=0)
        
        elif isinstance(data_a, tuple) and isinstance(data_b, tuple):
            if len(data_a) != 2 or len(data_b) != 2:
                raise ValueError("Both inputs should be tuples of length 2, (images, labels).")
            
            images_a, labels_a = data_a
            images_b, labels_b = data_b
            
            if images_a.shape[1:] != images_b.shape[1:]:
                raise ValueError(f"Shape mismatch in images: images_a has shape {images_a.shape} and images_b has shape {images_b.shape}. "
                                 f"Both must have the same shape except for the first dimension.")
            
            if labels_a.shape[1:] != labels_b.shape[1:]:
                raise ValueError(f"Shape mismatch in labels: labels_a has shape {labels_a.shape} and labels_b has shape {labels_b.shape}. "
                                 f"Both must have the same shape except for the first dimension.")
            
            merged_images = np.concatenate([images_a, images_b], axis=0)
            merged_labels = np.concatenate([labels_a, labels_b], axis=0)
            
            return (merged_images, merged_labels)
        
        else:
            raise TypeError("Inputs must be either both ndarrays or both tuples of (images, labels).")

    @staticmethod
    def getInputShape(data: Union[np.ndarray, pd.DataFrame, DirectoryIterator, NumpyArrayIterator]) -> tuple:
        if isinstance(data, NumpyArrayIterator):
            return data.x.shape[1:]

        if isinstance(data, DirectoryIterator):
            return (data.target_size[0], data.target_size[1], 3)

        if ((isinstance(data, np.ndarray) and data.ndim < 3) or isinstance(data, pd.DataFrame)):
            temp_data = data
            temp_data = pd.DataFrame(temp_data)
            return (temp_data.shape[1], )
        
        else:
            shape = data[0].shape
            if len(shape) == 2: 
                return (shape[0], shape[1], 1)
            else: 
                return (shape[0], shape[1], 3)   

    @staticmethod
    def getImageShape(image: Union[np.ndarray, Image.Image, str]) -> Tuple[int, int]:
        '''Provide the image or path to the image and get its dimensions size i.e. (32, 32).
        '''

        if isinstance(image, str):
            image = Image.open(image)

        if isinstance(image, Image.Image):
            image = np.array(image)

        if isinstance(image, np.ndarray):
            if image.ndim == 2:
                return (image.shape[0], image.shape[1])
            elif image.ndim == 3:
                return (image.shape[0], image.shape[1])
        
        raise ValueError("The image must be one of the following types: ``np.ndarray``, ``Image.Image`` or a ``str`` representing the path.")

    @staticmethod
    def ensureOneHotEncoding(
                y: Union[np.ndarray, pd.Series]
                ) -> np.ndarray:

        if isinstance(y, pd.Series):
            arr = y.astype('category').cat.codes
            return to_categorical(arr)

        if len(y.shape) > 1 and y.shape[1] > 1:
            return y

        return to_categorical(y)
    
    @staticmethod
    def datasetToArray(
            x: Union[np.ndarray, pd.DataFrame], 
            y: Union[np.ndarray, pd.Series]
            ) -> tuple[np.ndarray, np.ndarray]:
        
        new_y = FlashPreProcessing.ensureOneHotEncoding(y)
        new_x = x

        if isinstance(new_x, pd.DataFrame):
            new_x = new_x.values  # Convert DataFrame to NumPy array

        if isinstance(new_y, pd.Series):
            new_y = new_y.to_numpy()  # Convert Series to NumPy array
        
        return new_x, new_y
    
    @staticmethod
    def convertNdArrayToGrayScale(images: np.ndarray) -> np.ndarray:
        if images.shape[-1] == 1:
            return images
        grayscale_images = np.dot(images[...,:3], [0.2989, 0.5870, 0.1140])
        return np.expand_dims(grayscale_images, axis=-1)

    @staticmethod
    def resizeNpArray(array: np.ndarray, new_size1: int, new_size2: int) -> np.ndarray:
        if array.shape[1:3] == (new_size1, new_size2):
            return array
        if len(array.shape) == 3:
            array = np.expand_dims(array, axis=-1)
        resized_array = tf.image.resize(array, (new_size1, new_size2)).numpy()
        
        return resized_array

    @staticmethod
    def convertNdArrayToRGB(images: np.ndarray) -> np.ndarray:
        """
            Convert a batch of grayscale images (shape: (num_images, height, width) or 
            (num_images, height, width, 1)) into RGB format (shape: (num_images, height, width, 3)).
            
            Parameters:
            images: np.ndarray, Input image batch of shape (num_images, height, width) or 
                    (num_images, height, width, 1)
            
            Returns:
            np.ndarray, A batch of images with shape (num_images, height, width, 3).
        """

        if images.shape[-1] == 3:
            return images
        
        if len(images.shape) == 3:
            images = np.expand_dims(images, axis=-1)
            
        return np.repeat(images, 3, axis=-1)
    
    @staticmethod
    def minMaxScaler(x: pd.DataFrame | np.ndarray, min: float = 0, max: float = 1) -> np.ndarray:
        scaler = MinMaxScaler((min, max))
        return scaler.fit_transform(x)
    
    @staticmethod
    def labelEncoder(y: Union[np.ndarray, pd.Series, list], 
                     return_encoder: bool = False
                     ) -> np.ndarray | tuple[np.ndarray, LabelEncoder]:
        
        le: LabelEncoder = LabelEncoder()
        le.fit(y)

        if return_encoder:
            return le.transform(y), le

        return le.transform(y)
    
    @staticmethod
    def labelDecoder(labels: Union[np.ndarray, pd.Series, list], 
                     encoder: LabelEncoder
                     ) -> np.ndarray:

        return (encoder.inverse_transform(labels))